@@ -1,16 +1,20 @@
 from __future__ import annotations
 
 from collections.abc import Collection, Iterator, Mapping, Sequence
-from functools import partial
 from typing import TYPE_CHECKING, Any, Literal, cast, overload
 
 import pyarrow as pa
 import pyarrow.compute as pc
 
 from narwhals._arrow.series import ArrowSeries
-from narwhals._arrow.utils import native_to_narwhals_dtype
+from narwhals._arrow.utils import (
+    arange,
+    concat_tables,
+    narwhals_to_native_dtype,
+    native_to_narwhals_dtype,
+    repeat,
+)
 from narwhals._compliant import EagerDataFrame
-from narwhals._expression_parsing import ExprKind
 from narwhals._utils import (
     Implementation,
     Version,
@@ -41,7 +45,6 @@ if TYPE_CHECKING:
     from narwhals._arrow.namespace import ArrowNamespace
     from narwhals._arrow.typing import (  # type: ignore[attr-defined]
         ChunkedArrayAny,
-        Mask,
         Order,
     )
     from narwhals._compliant.typing import CompliantDataFrameAny, CompliantLazyFrameAny
@@ -73,7 +76,6 @@ if TYPE_CHECKING:
         "right outer",
         "full outer",
     ]
-    PromoteOptions: TypeAlias = Literal["none", "default", "permissive"]
 
 
 class ArrowDataFrame(
@@ -118,15 +120,60 @@ class ArrowDataFrame(
         /,
         *,
         context: _LimitedContext,
-        schema: IntoSchema | None,
+        schema: IntoSchema | Mapping[str, DType | None] | None,
+    ) -> Self:
+        if not schema and not data:
+            return cls.from_native(pa.table({}), context=context)
+        if not schema:
+            return cls.from_native(pa.table(data), context=context)  # type: ignore[arg-type]
+        if not any(dtype is None for dtype in schema.values()):
+            from narwhals.schema import Schema
+
+            pa_schema = Schema(cast("IntoSchema", schema)).to_arrow()
+            if pa_schema and not data:
+                native = pa_schema.empty_table()
+            else:
+                native = pa.Table.from_pydict(data, schema=pa_schema)
+            return cls.from_native(native, context=context)
+        if context._implementation._backend_version() < (14,):
+            msg = "Passing `None` dtype in `from_dict` requires PyArrow>=14"
+            raise NotImplementedError(msg)
+        res = pa.table(
+            {
+                name: pa.chunked_array(  # type: ignore[misc]
+                    [data[name] if data else []],
+                    type=narwhals_to_native_dtype(nw_dtype, version=context._version)
+                    if nw_dtype is not None
+                    else None,
+                )
+                for name, nw_dtype in schema.items()
+            }
+        )
+        return cls.from_native(pa.table(res), context=context)
+
+    @classmethod
+    def from_dicts(
+        cls,
+        data: Sequence[Mapping[str, Any]],
+        /,
+        *,
+        context: _LimitedContext,
+        schema: IntoSchema | Mapping[str, DType | None] | None,
     ) -> Self:
         from narwhals.schema import Schema
 
-        pa_schema = Schema(schema).to_arrow() if schema is not None else schema
+        if schema and any(dtype is None for dtype in schema.values()):
+            msg = "`from_dicts` with `schema` where any dtype is `None` is not supported for PyArrow."
+            raise NotImplementedError(msg)
+        pa_schema = (
+            Schema(cast("IntoSchema", schema)).to_arrow()
+            if schema is not None
+            else schema
+        )
         if pa_schema and not data:
             native = pa_schema.empty_table()
         else:
-            native = pa.Table.from_pydict(data, schema=pa_schema)
+            native = pa.Table.from_pylist(data, schema=pa_schema)
         return cls.from_native(native, context=context)
 
     @staticmethod
@@ -315,7 +362,7 @@ class ArrowDataFrame(
         )
 
     def select(self, *exprs: ArrowExpr) -> Self:
-        new_series = self._evaluate_into_exprs(*exprs)
+        new_series = self._evaluate_exprs(*exprs)
         if not new_series:
             # return empty dataframe, like Polars does
             return self._with_native(
@@ -342,7 +389,7 @@ class ArrowDataFrame(
         # NOTE: We use a faux-mutable variable and repeatedly "overwrite" (native_frame)
         # All `pyarrow` data is immutable, so this is fine
         native_frame = self.native
-        new_columns = self._evaluate_into_exprs(*exprs)
+        new_columns = self._evaluate_exprs(*exprs)
         columns = self.columns
 
         for col_value in new_columns:
@@ -387,12 +434,10 @@ class ArrowDataFrame(
             )
 
             return self._with_native(
-                self.with_columns(
-                    plx.lit(0, None).alias(key_token).broadcast(ExprKind.LITERAL)
-                )
+                self.with_columns(plx.lit(0, None).alias(key_token).broadcast())
                 .native.join(
                     other.with_columns(
-                        plx.lit(0, None).alias(key_token).broadcast(ExprKind.LITERAL)
+                        plx.lit(0, None).alias(key_token).broadcast()
                     ).native,
                     keys=key_token,
                     right_keys=key_token,
@@ -488,24 +533,21 @@ class ArrowDataFrame(
 
     def with_row_index(self, name: str, order_by: Sequence[str] | None) -> Self:
         plx = self.__narwhals_namespace__()
+        data = arange(0, len(self), 1)
         if order_by is None:
-            import numpy as np  # ignore-banned-import
-
-            data = pa.array(np.arange(len(self), dtype=np.int64))
             row_index = plx._expr._from_series(
                 plx._series.from_iterable(data, context=self, name=name)
             )
+            return self.select(row_index, plx.all())
+        indices = pc.sort_indices(self.native, [(by, "ascending") for by in order_by])
+        if self._backend_version < (20,):
+            new_col = data.take(pc.sort_indices(indices))
         else:
-            rank = plx.col(order_by[0]).rank("ordinal", descending=False)
-            row_index = (rank.over(partition_by=[], order_by=order_by) - 1).alias(name)
-        return self.select(row_index, plx.all())
+            new_col = pc.scatter(data, indices.cast(pa.int64()))  # type: ignore[attr-defined]
+        return self._with_native(self.native.add_column(0, name, new_col))
 
-    def filter(self, predicate: ArrowExpr | list[bool | None]) -> Self:
-        if isinstance(predicate, list):
-            mask_native: Mask | ChunkedArrayAny = predicate
-        else:
-            # `[0]` is safe as the predicate's expression only returns a single column
-            mask_native = self._evaluate_into_exprs(predicate)[0].native
+    def filter(self, predicate: ArrowExpr) -> Self:
+        mask_native = self._evaluate_single_output_expr(predicate).native
         return self._with_native(
             self.native.filter(mask_native), validate_column_names=False
         )
@@ -776,34 +818,19 @@ class ArrowDataFrame(
         variable_name: str,
         value_name: str,
     ) -> Self:
-        n_rows = len(self)
-        index_ = [] if index is None else index
-        on_ = [c for c in self.columns if c not in index_] if on is None else on
-        concat = (
-            partial(pa.concat_tables, promote_options="permissive")
-            if self._backend_version >= (14, 0, 0)
-            else pa.concat_tables
-        )
-        names = [*index_, variable_name, value_name]
-        return self._with_native(
-            concat(
-                [
-                    pa.Table.from_arrays(
-                        [
-                            *(self.native.column(idx_col) for idx_col in index_),
-                            cast(
-                                "ChunkedArrayAny",
-                                pa.array([on_col] * n_rows, pa.string()),
-                            ),
-                            self.native.column(on_col),
-                        ],
-                        names=names,
-                    )
-                    for on_col in on_
-                ]
-            )
-        )
         # TODO(Unassigned): Even with promote_options="permissive", pyarrow does not
         # upcast numeric to non-numeric (e.g. string) datatypes
+        n = len(self)
+        index = [] if index is None else list(index)
+        on_ = (c for c in self.columns if c not in index) if on is None else iter(on)
+        index_cols = self.native.select(index)
+        column = self.native.column
+        tables = (
+            index_cols.append_column(variable_name, repeat(name, n)).append_column(
+                value_name, column(name)
+            )
+            for name in on_
+        )
+        return self._with_native(concat_tables(tables, "permissive"))
 
     pivot = not_implemented()
